#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include "kd-tree.h"
#include "space.h"
#include "array.h"
#include "genbench.h"

#define DEBUG(s...)	fprintf(stderr, s)

kdtree_array *bank;
kdtree_space bank_space;

static void collision_handler(void *a, void *b)
{
	gb_collision((intptr_t)a, (intptr_t)b);
}

int main(int argc, char **argv, char **envp)
{
	kdtree_sphere bs;
	kdtree_iaabb ibb;
	int i, soobs;
	
	gb_init("Linear kd-tree sweep", argc, argv);
	
	bank = kdtree_array_new();

	kdtree_set_isospace(&bank_space,
		gb_min_x(), gb_max_x(),
		gb_min_y(), gb_max_y(),
		gb_min_z(), gb_max_z());
	

	while(1) {
		gb_timer_start(GB_COLLISION_PHASE);
		gb_timer_start(GB_POSITION_UPDATE);
		
		kdtree_clear(bank, gb_nobj());
		
		soobs = 0;
		for (i = 0; i < gb_nobj(); i++) {
			gb_obj_coord(i, &bs.x, &bs.y, &bs.z);
			bs.r = gb_obj_radius(i);
			
			if (kdtree_pack_sphere(&bank_space, &ibb, &bs)) {
				soobs++;
				kdtree_add_object(bank, NULL,
					(void *)(intptr_t)i);
			} else {
				kdtree_add_object(bank, &ibb,
					(void *)(intptr_t)i);
			}
		}
		gb_timer_stop(GB_POSITION_UPDATE);
		if (soobs) {
			DEBUG("%d out-of-bounds spheres placed at root\n",
				soobs);
		}

		gb_tell_nodes(bank->nobj_tree);

		kdtree_collisions(bank, &collision_handler);

		gb_timer_stop(GB_COLLISION_PHASE);

		/*{
			double svol, nvol;
			double volume = 0.0;
			kdtree_aabb bb;
			
			for (i = 0; i < bank->nobj_tree; i++) {
				kdtree_node_aabb(&bank_space,
					&bank->tree[i].path, &bb);
				nvol = (bb.x1 - bb.x0) * (bb.y1 - bb.y0)
					* (bb.z1 - bb.z0);
				volume += nvol;
			}
			
			svol = (4.0 / 3.0 * M_PI * gb_obj_radius(0)
				* gb_obj_radius(0) * gb_obj_radius(0));

			DEBUG("Spheres are %g; "
			"Node volume: Total %g, Average %g, "
			"Average ratio %12.4f. Nodes: %g N\n",
			svol, volume, volume / gb_nobj(),
			volume / (svol * gb_nobj()),
			(double)bank->nobj_tree / gb_nobj());
		}*/
		
		gb_next_step();
		//sleep(1);
	}

	return(0);
}


